import yaml
import pathspec
import os
from presidio_analyzer import AnalyzerEngine


class PresidioCLIConfigError(Exception):
    pass


class PresidioCLIConfig(object):
    def __init__(self, content=None, file=None):
        assert (content is None) ^ (file is None)
        self.ignore = None
        self.locale = None
        self.analyzer = AnalyzerEngine()
        self.threshold = 0
        self.language = "en"
        if file is not None:
            with open(file) as f:
                content = f.read()
        self.parse(content)
        self.validate()

    def is_file_ignored(self, filepath):
        return self.ignore and self.ignore.match_file(filepath)

    def is_text_file(self, filepath):
        """
        Detect if file is a not a binary file.
        Based on https://stackoverflow.com/a/7392391
        """

        textchars = bytearray(
            {7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7F}
        )
        with open(filepath, "rb") as f:
            # return true if it's not a binary
            return not bool(f.read(1024).translate(None, textchars))

    def extend(self, base_config):
        assert isinstance(base_config, PresidioCLIConfig)

        # Create list with unique entries
        if base_config.entities is not None:
            self.entities = list(set(base_config.entities + self.entities))

        if base_config.ignore is not None:
            self.ignore = base_config.ignore

        if base_config.language is not None:
            self.language = base_config.language

    def parse(self, raw_content):
        try:
            conf = yaml.safe_load(raw_content)
        except Exception as e:
            raise PresidioCLIConfigError("invalid config: %s" % e)

        if not isinstance(conf, dict):
            raise PresidioCLIConfigError("invalid config: not a dict")

        self.entities = conf.get("entities", {})

        if self.entities == {}:
            self.entities = self.analyzer.get_supported_entities()

        if "threshold" in conf:
            if not 0 <= float(self.threshold) <= 1:
                raise PresidioCLIConfigError(
                    f"Invalid threshold value: {self.threshold}. Threshold must be between 0 and 1"
                )
            self.threshold = float(conf["threshold"])
        if "language" in conf:
            self.language = conf["language"]
        if "extends" in conf:
            path = get_extended_config_file(conf["extends"])
            base = PresidioCLIConfig(file=path)
            try:
                self.extend(base)
            except Exception as e:
                raise PresidioCLIConfigError("invalid config: %s" % e)

        if "ignore" in conf:
            if not isinstance(conf["ignore"], str):
                raise PresidioCLIConfigError(
                    "invalid config: ignore should contain file patterns"
                )
            self.ignore = pathspec.PathSpec.from_lines(
                "gitwildmatch", conf["ignore"].splitlines()
            )

        if "locale" in conf:
            if not isinstance(conf["locale"], str):
                raise PresidioCLIConfigError(
                    "invalid config: locale should be a string"
                )
            self.locale = conf["locale"]

    def validate(self):
        for id in self.entities:
            try:
                assert id in self.analyzer.get_supported_entities()
            except Exception:
                raise PresidioCLIConfigError("invalid config: no such entity %s" % id)


def get_extended_config_file(name):
    # Is it a standard conf shipped with yamllint...
    if "/" not in name:
        std_conf = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "conf", name + ".yaml"
        )

        if os.path.isfile(std_conf):
            return std_conf

    # or a custom conf on filesystem?
    return name



import argparse
import os
import sys
import io
import locale
import platform
import json


from presidio_cli import SHELL_NAME, APP_DESCRIPTION, APP_VERSION
from presidio_cli.analyzer import analyze
from presidio_cli.config import PresidioCLIConfig, PresidioCLIConfigError


class Format(object):
    @staticmethod
    def parsable(problem):
        return json.dumps(problem.recognizer_result)

    @staticmethod
    def standard(problem):
        line = "  %d:%d" % (problem.line, problem.column)
        line += max(12 - len(line), 0) * " "
        line += str(problem.score)
        line += max(21 - len(line), 0) * " "
        line += problem.type
        if problem.explanation:
            line += "  (%s)" % problem.explanation
        return line

    @staticmethod
    def standard_color(problem):
        line = "  \033[2m%d:%d\033[0m" % (problem.line, problem.column)
        line += max(20 - len(line), 0) * " "
        if problem.score < 1:  # warning
            line += "\033[33m%s\033[0m" % problem.score
        else:
            line += "\033[31m%s\033[0m" % problem.score
        line += max(38 - len(line), 0) * " "
        line += problem.type
        if problem.explanation:
            line += "  \033[2m(%s)\033[0m" % problem.explanation
        return line

    @staticmethod
    def github(problem, filename):
        line = "::"
        line += str(problem.score)
        line += " file=" + filename + ","
        line += "line=" + format(problem.line) + ","
        line += "col=" + format(problem.column)
        line += "::"
        line += format(problem.line)
        line += ":"
        line += format(problem.column)
        line += " "
        line += "[" + problem.type + "] "
        if problem.explanation:
            line += problem.explanation
        return line


def supports_color():
    supported_platform = not (
        platform.system() == "Windows"
        and not (
            "ANSICON" in os.environ
            or ("TERM" in os.environ and os.environ["TERM"] == "ANSI")
        )
    )
    return supported_platform and hasattr(sys.stdout, "isatty") and sys.stdout.isatty()


def show_problems(problems, file, args_format, no_warn):
    max_level = 0
    first = True

    if args_format == "auto":
        if "GITHUB_ACTIONS" in os.environ and "GITHUB_WORKFLOW" in os.environ:
            args_format = "github"
        elif supports_color():
            args_format = "colored"

    for problem in problems:
        if no_warn and (problem.level != "error"):
            continue
        if args_format == "parsable":
            print(Format.parsable(problem))
        elif args_format == "github":
            if first:
                print("::group::%s" % file)
                first = False
            print(Format.github(problem, file))
        elif args_format == "colored":
            if first:
                print("\033[4m%s\033[0m" % file)
                first = False
            print(Format.standard_color(problem))
        else:
            if first:
                print(file)
                first = False
            print(Format.standard(problem))
        # max_level+=1

    if not first and args_format == "github":
        print("::endgroup::")

    if not first and args_format != "parsable":
        print("")

    return max_level


def find_files_recursively(items, conf):
    for item in items:
        if os.path.isdir(item):
            for root, dirnames, filenames in os.walk(item):
                for f in filenames:
                    filepath = os.path.join(root, f)
                    if conf.is_text_file(filepath):
                        yield filepath
        else:
            yield item


def run():
    parser = argparse.ArgumentParser(prog=SHELL_NAME, description=APP_DESCRIPTION)

    parser.add_argument("-v", "--version", action="version", version=f"v{APP_VERSION}")

    files_group = parser.add_mutually_exclusive_group(required=True)
    files_group.add_argument(
        "files",
        metavar="FILE_OR_DIR",
        nargs="*",
        default=(),
        help="files to check",
    )
    files_group.add_argument(
        "-", action="store_true", dest="stdin", help="read from standard input"
    )

    config_group = parser.add_mutually_exclusive_group()
    config_group.add_argument(
        "-c",
        "--config-file",
        dest="config_file",
        action="store",
        help="path to a custom configuration",
    )

    config_group.add_argument(
        "-d",
        "--config-data",
        dest="config_data",
        action="store",
        help="custom configuration (as YAML source)",
    )

    parser.add_argument(
        "-f",
        "--format",
        choices=("standard", "github", "auto", "colored", "parsable"),
        default="auto",
        help="format for parsing output",
    )

    parser.add_argument(
        "--no-warnings",
        action="store_true",
        help="output only error level problems",
    )

    args = parser.parse_args()

    try:
        if args.config_data is not None:
            if args.config_data != "" and ":" not in args.config_data:
                args.config_data = "extends: " + args.config_data
            conf = PresidioCLIConfig(content=args.config_data)
        elif args.config_file is not None:
            conf = PresidioCLIConfig(file=args.config_file)
        elif os.path.isfile(".presidiocli"):
            conf = PresidioCLIConfig(file=".presidiocli")
        else:
            conf = PresidioCLIConfig(content="extends: default")
    except PresidioCLIConfigError as e:
        print(e, file=sys.stderr)
        sys.exit(1)

    if conf.locale is not None:
        locale.setlocale(locale.LC_ALL, conf.locale)

    for file in find_files_recursively(args.files, conf):
        filepath = file[2:] if file.startswith("./") else file
        try:
            with io.open(file, newline="") as f:
                problems = analyze(f, conf, filepath)
        except EnvironmentError as e:
            print(e, file=sys.stderr)
            sys.exit(1)
        prob_num = show_problems(
            problems, file, args_format=args.format, no_warn=args.no_warnings
        )

    if args.stdin:
        try:
            problems = analyze(sys.stdin, conf, "")
        except EnvironmentError as e:
            print(e, file=sys.stderr)
            sys.exit(1)
        prob_num = show_problems(
            problems,
            "stdin",
            args_format=args.format,
            no_warn=args.no_warnings,
        )

    if prob_num > 0:
        return_code = 1
    else:
        return_code = 0
    sys.exit(return_code)



"""An python CLI for analyze PII Entities with Microsoft Presidio framework."""


APP_NAME = "presidio_cli"
APP_VERSION = "0.0.5"
APP_DESCRIPTION = __doc__
SHELL_NAME = "presidio"

__author__ = "Insights Engineering"
__author_email__ = "basel.nestcicd@roche.com"
__copyright__ = "Copyright 2022, F. Hoffman-La Roche Ltd"
__license__ = "Apache 2.0"
__version__ = APP_VERSION



# -*- coding: utf-8 -*-

from setuptools import setup
import os.path
from presidio_cli import (
    __author__,
    __author_email__,
    __license__,
    APP_NAME,
    APP_VERSION,
    APP_DESCRIPTION,
)

readme = ""
here = os.path.abspath(os.path.dirname(__file__))
readme_path = os.path.join(here, "README.md")
if os.path.exists(readme_path):
    with open(readme_path, "rb") as stream:
        readme = stream.read().decode("utf8")

setup(
    name=APP_NAME,
    version=APP_VERSION,
    author=__author__,
    author_email=__author_email__,
    url="https://github.com/insightsengineering/presidio-cli",
    license=__license__,
    description=APP_DESCRIPTION.split("\n")[0],
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=["presidio-analyzer>=2.2", "pyyaml", "pathspec"],
    trusted_host=["pypi.org"],
    tests_require=["pytest", "flake8>=3.7.9"],
)



from presidio_analyzer import RecognizerResult


class Line(object):
    def __init__(self, line_no, buffer, start, end):
        self.line_no = line_no
        self.start = start
        self.end = end
        self.buffer = buffer

    @property
    def content(self):
        return self.buffer[self.start : self.end]  # noqa


def line_generator(buffer):
    line_no = 1
    cur = 0
    next = buffer.find("\n")
    while next != -1:
        if next > 0 and buffer[next - 1] == "\r":
            yield Line(line_no, buffer, start=cur, end=next - 1)
        else:
            yield Line(line_no, buffer, start=cur, end=next)
        cur = next + 1
        next = buffer.find("\n", cur)
        line_no += 1

    yield Line(line_no, buffer, start=cur, end=len(buffer))


class PIIProblem(object):
    """Represents a PII problem found by presidio-cli."""

    def __init__(self, line, recognizer_result):
        assert isinstance(recognizer_result, RecognizerResult)
        self.recognizer_result = recognizer_result.to_dict()
        #: Line on which the problem was found (starting at 1)
        self.line = line
        #: Column on which the problem was found (starting at 1)
        self.column = self.recognizer_result["start"] + 1
        #: Human-readable description of the problem
        self.explanation = self.recognizer_result["analysis_explanation"]
        #: Identifier of the rule that detected the problem
        self.type = self.recognizer_result["entity_type"]
        # Score as a probability determined by the model
        self.score = self.recognizer_result["score"]


def _analyze(buffer, conf):
    """Analyze a text source.
    Returns a generator of PIIProblem objects.
    :param buffer: str, string to read from
    :param conf: presidio_cli configuration object
    """
    assert hasattr(
        buffer, "__getitem__"
    ), "_run() argument must be a buffer, not a stream"

    for line in line_generator(buffer):
        for result in conf.analyzer.analyze(
            text=line.content, entities=conf.entities, language=conf.language
        ):
            p = PIIProblem(line.line_no, result)
            if p.score >= conf.threshold:
                yield p


def analyze(input, conf, filepath=None):
    """Analyze a text source.
    Returns a generator of PIIProblem objects.
    :param input: buffer, string or stream to read from
    :param conf: presidio_cli configuration object
    :param filepath: string, string with path to file
    """

    if conf.is_file_ignored(filepath):
        return tuple()
    if isinstance(input, (bytes, str)):
        return _analyze(input, conf)
    elif hasattr(input, "read"):  # Python 2's file or Python 3's io.IOBase
        # We need to have everything in memory to parse correctly
        content = input.read()
        return _analyze(content, conf)
    else:
        raise TypeError("input should be a string or a stream")
